#!/usr/bin/perl -w

# mailgraph -- a postfix statistics rrdtool frontend
# copyright (c) 2000, 2001 David Schweikert <dws@ee.ethz.ch>
# released under the GNU General Public License

######## Parse::Syslog 0.03 (automatically embedded) ########
package Parse::Syslog;
use Carp;
use Symbol;
use Time::Local;
use strict;
use vars qw($VERSION);
my %months_map = (
    'Jan' => 0, 'Feb' => 1, 'Mar' => 2,
    'Apr' => 3, 'May' => 4, 'Jun' => 5,
    'Jul' => 6, 'Aug' => 7, 'Sep' => 8,
    'Oct' => 9, 'Nov' =>10, 'Dec' =>11
);
sub new($$;%)
{
    my ($class, $file, %data) = @_;
    croak "new() requires one argument: file" unless defined $file;
    %data = () unless %data;
    if(not defined $data{year}) {
        $data{year} = (localtime(time))[5]+1900;
    }
    $data{_repeat}=0;
    if(ref $file eq 'File::Tail') {
        $data{filetail} = 1;
        $data{file} = $file;
    }
    else {
        $data{file}=gensym;
        open($data{file}, "<$file") or croak "can't open $file: $!";
    }
    return bless \%data, $class;
}
sub _next_line($)
{
    my $self = shift;
    my $f = $self->{file};
    if(defined $self->{filetail}) {
        return $f->read;
    }
    else {
        return <$f>;
    }
}
sub next($)
{
    my ($self) = @_;
    while($self->{_repeat}>0) {
        $self->{_repeat}--;
        return $self->{_last_data};
    }
    line: while(my $str = $self->_next_line()) {
        # date, time and host 
        $str =~ /^
            (\w{3})\s+(\d+)   # date  -- 1, 2
            \s
            (\d+):(\d+):(\d+) # time  -- 3, 4, 5
            \s
            ([-\w\.]+)        # host  -- 6
            \s+
            (.*)              # text  -- 7
            $/x or do
        {
            carp "line not in syslog format: $str";
            next line;
        };
        my $mon = $months_map{$1};
        defined $mon or croak "unknown month $1\n";
        # year change
        $self->{year}++ if defined $self->{_last_mon}
            and $self->{_last_mon} > $mon;
        $self->{_last_mon} = $mon;
        # convert to unix time
        my $time;
        if($self->{GMT}) {
            $time = timegm($5,$4,$3,$2,$mon,$self->{year}-1900);
        }
        else {
            $time = timelocal($5,$4,$3,$2,$mon,$self->{year}-1900);
        }
        my ($host, $text) = ($6, $7);
        # last message repeated ... times
        if($text =~ /^last message repeated (\d+) time/) {
            next line if defined $self->{repeat} and not $self->{repeat};
            next line if not defined $self->{_last_data};
            $1 > 0 or do {
                carp "last message repeated 0 or less times??";
                next line;
            };
            $self->{_repeat}=$1-1;
            return $self->{_last_data};
        }
        # some systems send over the network their
        # hostname prefixed to the text. strip that.
        $text =~ s/^[-\w+\.]+\s+//;
        # snort log
        $text =~ /^
            (\S+?)          # program   -- 1
            (?:\[(\d+)\])?  # PID       -- 2
            :\s+
            (?:\[ID\ (\d+)\ ([a-z0-9]+)\.([a-z]+)\]\ )?   # Solaris 8 "message id" -- 3, 4, 5
            (.*)            # text      -- 6
            $/x or do
        {
            carp "line not in syslog format: $str";
            next line;
        };
        $self->{_last_data} = {
            timestamp => $time,
            host      => $host,
            program   => $1,
            pid       => $2,
            msgid     => $3,
            facility  => $4,
            level     => $5,
            text      => $6,
        };
        return $self->{_last_data};
    }
    return undef;
}

#####################################################################
#####################################################################
#####################################################################

use RRDs;

use strict;
use File::Tail;
use Getopt::Long;

my $VERSION = 0.15;

# config
my $rrdstep = 60;
my $xpoints = 540;
my $points_per_sample = 3;

# global variables
my $logfile;
my $rrd = "mailgraph.rrd";
my $year;
my $this_minute;
my %sum = ( sent => 0, received => 0, bounced => 0, rejected => 0 );
my $rrd_inited=0;

# prototypes
sub process_line($);
sub event_sent($);
sub event_received($);
sub event_bounced($);
sub event_rejected($);
sub init_rrd($);
sub update($);

sub usage
{
	print "usage: mailgraph [*options*]\n\n";
	print "  -c, --cat          causes the logfile to be only read and not monitored\n";
	print "  -l, --logfile f    monitor logfile f instead of /var/log/syslog\n";
	print "  -h, --help         display this help and exit\n";
	print "  -v, --version      output version information and exit\n";

	exit;
}

sub main
{
	my %opt = ();
	GetOptions(\%opt, 'help|h', 'cat|c', 'logfile|l=s', 'version|v') or usage;
	usage if $opt{help};

	if($opt{version}) {
		print "mailgraph $VERSION by dws\@ee.ethz.ch\n";
		exit;
	}

	my $logfile = defined $opt{logfile} ? $opt{logfile} : '/var/log/syslog';
	my $file;
	if($opt{cat}) {
		$file = $logfile;
	}
	else {
		$file = File::Tail->new(name=>$logfile, tail=>-1);
	}
	my $parser = new Parse::Syslog($file);
	while(my $sl = $parser->next) {
		process_line($sl);
	}
}

sub init_rrd($)
{
	my $m = shift;
	if(not -f $rrd) {
		my $rows = $xpoints/$points_per_sample;
		my $realrows = int($rows*1.1); # ensure that the full range is covered
		my $day_steps = int(3600*24 / ($rrdstep*$rows));
		# use multiples, otherwise rrdtool could choose the wrong RRA
		my $week_steps = $day_steps*7;
		my $month_steps = $week_steps*5;
		my $year_steps = $month_steps*12;
		RRDs::create($rrd, '--start', $m, '--step', $rrdstep,
				'DS:sent:ABSOLUTE:'.($rrdstep*2).':0:U',
				'DS:recv:ABSOLUTE:'.($rrdstep*2).':0:U',
				'DS:bounced:ABSOLUTE:'.($rrdstep*2).':0:U',
				'DS:rejected:ABSOLUTE:'.($rrdstep*2).':0:U',
				"RRA:AVERAGE:0.5:$day_steps:$realrows",   # day
				"RRA:AVERAGE:0.5:$week_steps:$realrows",  # week
				"RRA:AVERAGE:0.5:$month_steps:$realrows", # month
				"RRA:AVERAGE:0.5:$year_steps:$realrows",  # year
				"RRA:MAX:0.5:$day_steps:$realrows",   # day
				"RRA:MAX:0.5:$week_steps:$realrows",  # week
				"RRA:MAX:0.5:$month_steps:$realrows", # month
				"RRA:MAX:0.5:$year_steps:$realrows",  # year
				);
		$this_minute = $m;
	}
	else {
		$this_minute = RRDs::last($rrd) + $rrdstep;
	}
	$rrd_inited=1;
}

sub process_line($)
{
	my $sl = shift;

	$sl->{program} =~ /^postfix\/(.*)/ or return;
	my $prog = $1;
	my $time = $sl->{timestamp};
	my $text = $sl->{text};
	if($prog eq 'smtp') {
		if($text =~ /\bstatus=sent\b/) {
			event_sent($time);
		}
		if($text =~ /\bstatus=bounced\b/) {
			event_bounced($time);
		}
	}
	elsif($prog eq 'local') {
		if($text =~ /\bstatus=bounced\b/) {
			event_bounced($time);
		}
	}
	elsif($prog eq 'smtpd') {
		if($text =~ /^[0-9A-F]+: client=/) {
			event_received($time);
		}
		if($text =~ /^reject: /) {
			event_rejected($time);
		}
	}
}

sub event_sent($)
{
	my $t = shift;
	update($t) and $sum{sent}++;
}

sub event_received($)
{
	my $t = shift;
	update($t) and $sum{received}++;
}

sub event_bounced($)
{
	my $t = shift;
	update($t) and $sum{bounced}++;
}

sub event_rejected($)
{
	my $t = shift;
	update($t) and $sum{rejected}++;
}

# returns 1 if $sum should be updated
sub update($)
{
	my $t = shift;
	my $m = $t - $t%$rrdstep;
	init_rrd($m) unless $rrd_inited;
	return 1 if $m == $this_minute;
	return 0 if $m < $this_minute;

	#print "$m\n";
	print "update $this_minute:$sum{sent}:$sum{received}:$sum{bounced}:$sum{rejected}\n";
	RRDs::update $rrd, "$this_minute:$sum{sent}:$sum{received}:$sum{bounced}:$sum{rejected}";
	if($m > $this_minute+$rrdstep) {
		for(my $sm=$this_minute+$rrdstep;$sm<$m;$sm+=$rrdstep) {
			print "update $sm:0:0:0:0 (SKIP)\n";
			RRDs::update $rrd, "$sm:0:0:0:0";
		}
	}
	$this_minute = $m;
	$sum{sent}=0;
	$sum{received}=0;
	$sum{bounced}=0;
	$sum{rejected}=0;
	return 1;
}

main;
